#  ----------------------------------------------------------------------------
#  "THE BEER-WARE LICENSE" (Revision 42):
#  dkratzert@gmx.de> wrote this file.  As long as you retain
#  this notice you can do whatever you want with this stuff. If we meet some day,
#  and you think this stuff is worth it, you can buy me a beer in return.
#  Dr. Daniel Kratzert
#  ----------------------------------------------------------------------------
#
# Settings are stored as a single JSON file in the user's config directory.
# On first launch after the migration, existing QSettings data (Windows registry /
# INI file / macOS plist) is automatically imported into the JSON file.
# The JSON structure uses nested dicts where top-level keys correspond to the
# former QSettings groups (e.g. "MainWindow", "equipment", "property", ...).
# Ungrouped keys sit at the root level.

from __future__ import annotations

import json
import logging
import os
import threading
import time
from collections.abc import Iterable
from dataclasses import asdict, fields
from pathlib import Path
from typing import Any

logger = logging.getLogger(__name__)

# ---------------------------------------------------------------------------
# Helpers: default config directory
# ---------------------------------------------------------------------------

def _default_config_dir() -> Path:
    """Return the platform-appropriate config directory for FinalCif."""
    if os.name == 'nt':
        base = Path(os.environ.get('APPDATA', Path.home()))
    elif os.name == 'posix' and 'darwin' in os.uname().sysname.lower():
        base = Path.home() / 'Library' / 'Application Support'
    else:
        base = Path(os.environ.get('XDG_CONFIG_HOME', Path.home() / '.config'))
    return base / 'FinalCif'


_SETTINGS_FILENAME = 'settings.json'


# ---------------------------------------------------------------------------
# JSON encoder / decoder for special types
# ---------------------------------------------------------------------------

def _custom_encoder(obj: Any) -> Any:
    """JSON encoder hook for Qt types, Author dataclass, and enums."""
    # Import lazily so the settings module doesn't hard-depend on Qt at import time
    # when only plain types are stored.
    from finalcif.equip_property.author_loop_templates import Author, AuthorType

    if hasattr(obj, '__class__') and obj.__class__.__name__ == 'QPoint':
        return {'__type__': 'QPoint', 'x': obj.x(), 'y': obj.y()}
    if hasattr(obj, '__class__') and obj.__class__.__name__ == 'QSize':
        return {'__type__': 'QSize', 'width': obj.width(), 'height': obj.height()}
    if isinstance(obj, Author):
        d = asdict(obj)
        d['__type__'] = 'Author'
        d['author_type'] = obj.author_type.value if isinstance(obj.author_type, AuthorType) else str(obj.author_type)
        return d
    if isinstance(obj, AuthorType):
        return obj.value
    raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


def _custom_decoder(d: dict) -> Any:
    """JSON object-hook that reconstitutes tagged dicts back to rich types."""
    if '__type__' not in d:
        return d

    t = d['__type__']
    if t == 'QPoint':
        from qtpy.QtCore import QPoint
        return QPoint(d['x'], d['y'])
    if t == 'QSize':
        from qtpy.QtCore import QSize
        return QSize(d['width'], d['height'])
    if t == 'Author':
        from finalcif.equip_property.author_loop_templates import Author, AuthorType
        kw = {f.name: d.get(f.name) for f in fields(Author)}
        at = kw.get('author_type')
        if isinstance(at, str):
            kw['author_type'] = AuthorType(at)
        return Author(**kw)
    return d


# ---------------------------------------------------------------------------
# Template file I/O (used by export/import all templates)
# ---------------------------------------------------------------------------

def load_template_file(filepath: Path) -> dict | None:
    """Load a template file, trying JSON first, then legacy pickle.

    Parameters
    ----------
    filepath : Path
        Path to a ``.json`` or legacy ``.dat`` template file.

    Returns
    -------
    dict | None
        The template data dictionary, or *None* if the file is empty.
    """
    # New JSON format
    if filepath.suffix.lower() == '.json':
        with open(filepath, 'r', encoding='utf-8') as f:
            return json.load(f, object_hook=_custom_decoder)
    # For .dat files, try JSON first then fall back to legacy pickle
    raw = filepath.read_bytes()
    stripped = raw.lstrip()
    if stripped and stripped[:1] in (b'{', b'['):
        return json.loads(raw.decode('utf-8'), object_hook=_custom_decoder)
    # Legacy pickle fallback
    import pickle  # only used to read legacy .dat files
    logger.info("Loading legacy pickle template file: %s", filepath)
    with open(filepath, 'rb') as f:
        return pickle.load(f)


# ---------------------------------------------------------------------------
# One-time QSettings -> JSON migration
# ---------------------------------------------------------------------------

def _migrate_qsettings_to_json(json_path: Path) -> dict:
    """Read all data from the existing QSettings store and return it as a dict.

    The resulting dict is written to *json_path* by the caller.  The original
    QSettings data is **not** deleted so that a downgrade is still possible.
    """
    from qtpy.QtCore import QSettings

    qs = QSettings('DK', 'FinalCif')
    data: dict[str, Any] = {'_migrated_from_qsettings': True}

    # Collect top-level keys (outside any group)
    for key in qs.allKeys():
        if '/' not in key:
            data[key] = _convert_qsettings_value(qs.value(key))

    # Collect grouped keys
    for group in qs.childGroups():
        qs.beginGroup(group)
        group_dict: dict[str, Any] = {}
        for key in qs.allKeys():
            group_dict[key] = _convert_qsettings_value(qs.value(key))
        qs.endGroup()
        data[group] = group_dict

    return data


def _convert_qsettings_value(value: Any) -> Any:  # noqa: PLR0911
    """Best-effort conversion of a QSettings value to a JSON-safe Python type."""
    from finalcif.equip_property.author_loop_templates import Author, AuthorType

    if value is None:
        return None

    # QPoint / QSize
    cls_name = type(value).__name__
    if cls_name == 'QPoint':
        return {'__type__': 'QPoint', 'x': value.x(), 'y': value.y()}
    if cls_name == 'QSize':
        return {'__type__': 'QSize', 'width': value.width(), 'height': value.height()}
    if isinstance(value, Author):
        return _custom_encoder(value)
    if isinstance(value, AuthorType):
        return value.value

    # Recurse into containers
    if isinstance(value, dict):
        return {str(k): _convert_qsettings_value(v) for k, v in value.items()}
    if isinstance(value, list | tuple):
        return [_convert_qsettings_value(v) for v in value]

    # Primitive types that JSON handles natively
    if isinstance(value, str | int | float | bool):
        return value

    # Fallback - store as string
    return str(value)


# ---------------------------------------------------------------------------
# Core JSON-backed settings class
# ---------------------------------------------------------------------------

class FinalCifSettings:
    """Application settings stored as a JSON file.

    Parameters
    ----------
    settings_path : Path | None
        Explicit path to the JSON file.  When *None* the platform default is
        used.  Pass an explicit path in tests for isolation.
    """

    def __init__(self, settings_path: Path | None = None):
        self.software_name = 'FinalCif'
        self.organization = 'DK'

        if settings_path is not None:
            self._path = settings_path
        else:
            self._path = _default_config_dir() / _SETTINGS_FILENAME

        self._data: dict[str, Any] = self._load_json()
        self._pending_timer: threading.Timer | None = None
        self._timer_lock = threading.Lock()

        # If the JSON file was empty / missing, attempt a one-time migration
        if not self._data:
            try:
                self._data = _migrate_qsettings_to_json(self._path)
                if self._data:
                    self._flush()
                    logger.warning("Migrated QSettings to %s", self._path)
            except Exception:
                logger.warning("No QSettings data to migrate (or Qt not available).", exc_info=True)
                self._data = {}

        # Cached property helpers (same semantics as before)
        self.property_keys_and_values = self.load_property_keys_and_values()
        self.property_keys = self.load_cif_keys_of_properties()

    # ------------------------------------------------------------------
    # Low-level JSON I/O
    # ------------------------------------------------------------------

    def _load_json(self) -> dict:
        if self._path.exists():
            try:
                with open(self._path, encoding='utf-8') as fh:
                    return json.load(fh, object_hook=_custom_decoder)
            except (json.JSONDecodeError, OSError):
                logger.warning("Could not read %s - starting with empty settings.", self._path, exc_info=True)
        return {}

    def _flush(self) -> None:
        """Atomically write the in-memory dict to disk (best-effort, fail-soft)."""
        try:
            self._path.parent.mkdir(parents=True, exist_ok=True)
            tmp = self._path.with_suffix('.tmp')
            with open(tmp, 'w', encoding='utf-8') as fh:
                json.dump(self._data, fh, default=_custom_encoder, ensure_ascii=False, indent=2)
            os.replace(tmp, self._path)
        except OSError as exc:
            logger.warning("Settings flush failed (will retry on close): %s", exc)

    def _schedule_flush(self) -> None:
        """Coalesce rapid writes: the actual disk I/O happens after 250 ms of inactivity."""
        with self._timer_lock:
            if self._pending_timer is not None:
                self._pending_timer.cancel()
            self._pending_timer = threading.Timer(0.25, self._flush)
            self._pending_timer.daemon = True
            self._pending_timer.start()

    def flush(self, retries: int = 3, delay: float = 0.05) -> None:
        """Persist in-memory data to disk immediately, retrying on failure.

        Call this on the shutdown path so the last write is not lost.
        Cancels any pending debounced flush first.
        """
        with self._timer_lock:
            if self._pending_timer is not None:
                self._pending_timer.cancel()
                self._pending_timer = None
        for attempt in range(retries):
            try:
                self._path.parent.mkdir(parents=True, exist_ok=True)
                tmp = self._path.with_suffix('.tmp')
                with open(tmp, 'w', encoding='utf-8') as fh:
                    json.dump(self._data, fh, default=_custom_encoder, ensure_ascii=False, indent=2)
                os.replace(tmp, self._path)
                return
            except OSError as exc:
                logger.warning("Settings flush attempt %d/%d failed: %s", attempt + 1, retries, exc)
                if attempt < retries - 1:
                    time.sleep(delay)
        logger.error("Settings could not be persisted after %d attempts.", retries)

    # ------------------------------------------------------------------
    # Internal helpers (replace beginGroup / endGroup pattern)
    # ------------------------------------------------------------------

    def _get_group(self, group: str) -> dict:
        """Return the sub-dict for *group*, creating it if needed."""
        return self._data.setdefault(group, {})

    def _set_in_group(self, group: str, key: str, value: Any) -> None:
        self._get_group(group)[key] = value
        self._schedule_flush()

    def _get_from_group(self, group: str, key: str, default: Any = None) -> Any:
        return self._get_group(group).get(key, default)

    # ------------------------------------------------------------------
    # Public API  (unchanged signatures)
    # ------------------------------------------------------------------

    @property
    def property_items(self):
        return self.list_saved_items(property='property')

    def save_window_position(self, position: Any, size: Any, maximized: bool) -> None:
        """Save main-window geometry.  *position* and *size* may be QPoint/QSize or plain dicts."""
        grp = self._get_group('MainWindow')
        grp['position'] = position
        grp['size'] = size
        grp['maximized'] = bool(maximized)
        self._schedule_flush()

    def load_window_position(self) -> dict:
        """Load window position information with sensible defaults."""
        from qtpy.QtCore import QPoint, QSize

        grp = self._get_group('MainWindow')
        pos = grp.get('position')
        size = grp.get('size')

        # Ensure we return QPoint / QSize for the caller
        if isinstance(size, dict) and '__type__' in size:
            size = _custom_decoder(size)
        if isinstance(pos, dict) and '__type__' in pos:
            pos = _custom_decoder(pos)

        if not isinstance(size, QSize) or size.width() <= 0:
            size = QSize(900, 850)
        if not isinstance(pos, QPoint) or pos.x() <= 0:
            pos = QPoint(20, 20)

        maximized = grp.get('maximized', False)
        if isinstance(maximized, str):
            maximized = maximized.lower() in ('true', '1', 'yes')

        return {'size': size, 'position': pos, 'maximized': bool(maximized)}

    def save_current_dir(self, directory: str) -> None:
        """Save the current work directory."""
        self._set_in_group('WorkDir', 'dir', directory)

    def load_last_workdir(self) -> str:
        return self._get_from_group('WorkDir', 'dir', '') or ''

    def get_equipment_list(self) -> list:
        return sorted(self.list_saved_items('equipment'))

    def get_properties_list(self) -> list:
        return sorted(self.list_saved_items('property'))

    def load_property_keys_and_values(self) -> list:
        keylist = []
        for p in self.property_items:
            try:
                keylist.append(self.load_settings_list(property='property', item_name=p))
            except TypeError:
                pass
        return keylist

    def load_cif_keys_of_properties(self) -> list[str]:
        return [x[0] for x in self.load_property_keys_and_values()]

    def load_property_values_by_key(self, cif_key: str):
        num_value_pairs = []
        if self.property_keys and cif_key in self.property_keys:
            keys_and_values = self.property_keys_and_values[self.property_keys.index(cif_key)]
            if len(keys_and_values) >= 1:
                property_values = keys_and_values[1]
            else:
                property_values = ['']
            for n, val in enumerate(property_values):
                num_value_pairs.append((n, val))
            return num_value_pairs
        return [(0, '')]

    def save_template_list(self, name: str, items: list):
        """Save a template list at the root level."""
        logger.debug("Saving %s %s", name, items)
        self._data[name] = items
        self._schedule_flush()
        self.property_keys_and_values = self.load_property_keys_and_values()
        self.property_keys = self.load_cif_keys_of_properties()

    def save_key_value(self, name: str, item: str | list | tuple | dict | int | bool):
        """Save a single key/value pair at the root level."""
        self._data[name] = item
        self._schedule_flush()
        logger.debug("Saving %s %s", name, item)

    def load_value_of_key(self, key: str) -> Iterable | list | int | float | None:
        """Load a root-level value."""
        return self._data.get(key)

    def delete_template(self, property: str, name: str):
        """Delete a template entry from a group."""
        grp = self._get_group(property)
        grp.pop(name, None)
        self._schedule_flush()
        if property == 'equipment':
            deleted = self.load_value_of_key(key='deleted_templates') or []
            deleted.append(name)
            deleted = list(set(deleted))
            self.save_key_value(name='deleted_templates', item=deleted)

    @property
    def deleted_equipment(self):
        deleted = self.load_value_of_key(key='deleted_templates')
        return deleted or []

    def empty_deleted_list(self):
        self.save_key_value(name='deleted_templates', item=[])

    def load_options(self) -> dict:
        options = self.load_settings_dict('Options', 'options')
        if not options:
            options = {
                'report_text': True,
                'picture_width': 7.5,
                'without_h': False,
                'report_adp': True,
                'track_changes': False,
                'checkcif_url': 'https://checkcif.iucr.org/cgi-bin/checkcif_hkl.pl',
            }
        options.update({
            'atoms_table': True,
            'bonds_table': True,
            'hydrogen_bonds': True,
        })
        return options

    def save_options(self, options: dict):
        self.save_settings_dict('Options', 'options', options)

    def load_settings_dict(self, property: str = '', item_name: str = '') -> dict:
        settings = self._load_settings(property, item_name)
        if settings is None:
            return {}
        return settings

    def load_settings_list(self, property: str = '', item_name: str = '') -> list:
        settings = self._load_settings(property, item_name)
        if isinstance(settings, list):
            return settings
        return []

    def load_settings_list_as_dict(self, property: str, item_name: str):
        setting = self.load_settings_list(property, item_name)
        keydict = {}
        for p in setting:
            try:
                keydict[p[0]] = p[1]
            except (IndexError, TypeError):
                continue
        return keydict

    def _load_settings(self, property: str, item_name: str) -> Any | None:
        grp = self._get_group(property)
        if item_name in grp:
            return grp[item_name]
        return None

    def list_saved_items(self, property: str) -> list:
        return list(self._get_group(property).keys())

    def save_settings_dict(self, property: str, name: str, items) -> None:
        self._set_in_group(property, name, items)

    def _save_settings_value(self, items: Any, name: str, property: str) -> None:
        self._set_in_group(property, name, items)
        logger.debug("Saving %s %s", name, items)

    def save_settings_list(self, property: str, name: str, items: list):
        self._save_settings_value(items, name, property)

    # ------------------------------------------------------------------
    # New convenience wrappers (replace direct QSettings access)
    # ------------------------------------------------------------------

    def load_recent_files(self) -> list[str]:
        """Return the list of recently opened files."""
        val = self._data.get('recent_files')
        if isinstance(val, list):
            return list(val)
        return []

    def save_recent_files(self, recent: list[str]) -> None:
        """Persist the recent-files list."""
        self._data['recent_files'] = recent
        self._schedule_flush()

    def load_property_list(self) -> list | None:
        """Return the root-level property_list value (used by properties import)."""
        return self._data.get('property_list')


if __name__ == '__main__':
    s = FinalCifSettings()
    print('load_property_by_key:', s.load_property_values_by_key(cif_key='_diffrn_ambient_environment'))
    print(s.load_cif_keys_of_properties())



# Create a QTextedit widget with a button to open a file dialog
# to select a file to be inserted into the QTextedit widget
# The file dialog is opened by pressing the button.
# The button should be placed above the QTextedit widget and appear on mouse over.

import sys

from qtpy import QtWidgets, QtCore, QtGui


class FloatingButtonWidget(QtWidgets.QPushButton):
    floatingButtonClicked = QtCore.Signal()

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent=parent, *args, **kwargs)
        self.setText('edit')
        self.paddingLeft = 2
        self.paddingTop = 2
        self.setFixedSize(24, 16)
        self.setFlat(True)
        self.setStyleSheet(f'''QPushButton {{background-color: {QtGui.QColor(220, 232, 247).name()};
                                            padding-bottom: 1px;
                                            border-radius: 5px; 
                                            font-size: 11px; 
                                            border: 1px solid {QtGui.QColor(170, 172, 177).name()};
                                            }} 
                                            ''')

    def update_position(self):
        """Update the position of the button to be placed in the bottom right corner of the parent widget.
        """
        self.move(self.parent().width() - self.width() - self.paddingLeft,
                  self.parent().height() - self.height() - self.paddingTop)

    def mousePressEvent(self, event):
        self.floatingButtonClicked.emit()


class OverlayTestEdit(QtWidgets.QTextEdit):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.edit_button = FloatingButtonWidget(parent=self)
        # The button should be placed over the QTextedit widget as overlay and appear on mouse over.
        self.edit_button.hide()
        self.edit_button.update_position()

    def resizeEvent(self, event):
        self.edit_button.update_position()
        super().resizeEvent(event)

    def enterEvent(self, a0):
        print('enter')
        self.edit_button.show()

    def leaveEvent(self, a0):
        print('leave')
        self.edit_button.hide()


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    window = OverlayTestEdit()
    window.resize(130, 30)
    window.show()
    sys.exit(app.exec())



#   ----------------------------------------------------------------------------
#   "THE BEER-WARE LICENSE" (Revision 42):
#   Daniel Kratzert <dkratzert@gmx.de> wrote this file.  As long as you retain
#   this notice you can do whatever you want with this stuff. If we meet some day,
#   and you think this stuff is worth it, you can buy me a beer in return.
#   ----------------------------------------------------------------------------
import os
from unittest.mock import patch

os.environ["RUNNING_TEST"] = 'True'
import unittest
from tests.helpers import AppWindowTestCase
from pathlib import Path

from qtpy.QtCore import Qt, QEvent, QModelIndex
from qtpy.QtGui import QColor, QBrush, QKeyEvent
from qtpy.QtWidgets import QApplication

from finalcif.appwindow import AppWindow
from finalcif.gui.cif_table_model import CifTableModel, CifRowData
from finalcif.gui.cif_table_view import CifTableView
from finalcif.gui.custom_classes import Column
from finalcif.tools.misc import unify_line_endings

data = Path('tests')


# noinspection PyMissingTypeHints
class TestMainTableFieldBehavior(AppWindowTestCase):

    def setUp(self) -> None:
        self.testcif = (data / 'examples/1979688.cif').absolute()
        (data / '/examples/1979688-finalcif_changes.cif').unlink(missing_ok=True)
        self.app = AppWindow(file=self.testcif)
        # self.app.show()
        self.app.settings.empty_deleted_list()
        self.app.ui.trackChangesCifCheckBox.setChecked(False)

    def tearDown(self) -> None:
        self.app.cif.finalcif_file.unlink(missing_ok=True)
        self.app.close()
        (data / 'examples/1979688-finalcif_changes.cif').unlink(missing_ok=True)
        super().tearDown()

    def key_row(self, key: str) -> int:
        return self.app.ui.cif_main_table.row_from_key(key)

    def cell_widget(self, row: int, col: int) -> str:
        return str(self.app.ui.cif_main_table.cellWidget(row, col).__class__)

    def cell_text(self, key: str, col: int) -> str:
        return unify_line_endings(self.app.ui.cif_main_table.getTextFromKey(key, col))

    def equipment_click(self, field: str):
        self.app.ui.EquipmentTemplatesStackedWidget.setCurrentIndex(0)
        item = self.app.ui.EquipmentTemplatesListWidget.findItems(field, Qt.MatchStartsWith)[0]
        self.app.ui.EquipmentTemplatesListWidget.setCurrentItem(item)
        self.app.equipment.load_selected_equipment()

    ######

    def test_delete_row(self):
        rows = self.app.ui.cif_main_table.rowCount()
        self.app.ui.cif_main_table.delete_row(self.key_row('_audit_update_record'))
        self.assertEqual(rows - 1, self.app.ui.cif_main_table.rowCount())

    def test_get_text_by_key(self):
        self.assertEqual('geom', self.cell_text('_atom_sites_solution_hydrogens', Column.CIF))
        self.assertEqual('', self.cell_text('_atom_sites_solution_hydrogens', Column.DATA))
        self.assertEqual('', self.cell_text('_atom_sites_solution_hydrogens', Column.EDIT))

    def test_load_equipment(self):
        self.app.equipment.import_equipment_from_file(str(data.parent / 'test-data/Crystallographer_Details.cif'))
        # make sure contact author is selected
        self.equipment_click('Crystallographer Details')
        # It is important here, that the first column has 'dkratzert@gmx.de' in it:
        self.assertEqual('?',
                         self.app.ui.cif_main_table.getTextFromKey('_audit_contact_author_email', Column.CIF))
        self.assertEqual('dkratzert@gmx.de',
                         self.app.ui.cif_main_table.getTextFromKey('_audit_contact_author_email', Column.DATA))
        self.assertEqual('dkratzert@gmx.de',
                         self.app.ui.cif_main_table.getTextFromKey('_audit_contact_author_email', Column.EDIT))

    def test_field_types(self):
        # In the model/view architecture, all cells have persistent widgets (no QTableWidgetItem).
        self.assertEqual("<class 'finalcif.gui.plaintextedit.MyQPlainTextEdit'>",
                         str(self.app.ui.cif_main_table.widget_from_key('_atom_sites_solution_hydrogens',
                                                                          Column.CIF).__class__))
        self.assertEqual("<class 'finalcif.gui.combobox.MyComboBox'>",
                         str(self.app.ui.cif_main_table.widget_from_key('_atom_sites_solution_hydrogens',
                                                                          Column.EDIT).__class__))

    def test_multicif(self):
        self.assertEqual(False, self.app.cif.is_multi_cif)
        self.assertEqual(1, len(self.app.cif.doc))
        self.app.append_cif(data.parent / 'test-data/1000006.cif')
        self.assertEqual(True, self.app.cif.is_multi_cif)
        self.assertEqual(2, len(self.app.cif.doc))


class TestMainTableRowHeights(AppWindowTestCase):
    """Tests that row heights in the main CIF table are reasonable.

    Exact pixel values vary across platforms, fonts, and DPI settings,
    so all assertions use generous range checks.
    """

    def setUp(self) -> None:
        self.testcif = (data / 'examples/1979688.cif').absolute()
        (data / 'examples/1979688-finalcif_changes.cif').unlink(missing_ok=True)
        self.app = AppWindow(file=self.testcif)
        self.app.settings.empty_deleted_list()
        self.app.ui.trackChangesCifCheckBox.setChecked(False)
        self.table = self.app.ui.cif_main_table
        self.table.resizeRowsToContents()

    def tearDown(self) -> None:
        self.app.cif.finalcif_file.unlink(missing_ok=True)
        self.app.close()
        (data / 'examples/1979688-finalcif_changes.cif').unlink(missing_ok=True)
        super().tearDown()

    def test_single_line_row_height(self):
        """A short single-line value like 'geom' should produce a compact row."""
        row = self.table.row_from_key('_atom_sites_solution_hydrogens')
        h = self.table.rowHeight(row)
        self.assertGreaterEqual(h, 10)
        self.assertLessEqual(h, 80)

    def test_separator_row_is_short(self):
        """The blue diagonal separator row should be very compact."""
        sep_row = self.app.complete_data_row
        self.assertGreaterEqual(sep_row, 0, 'Separator row not found')
        h = self.table.rowHeight(sep_row)
        self.assertLessEqual(h, 40)

    def test_shelx_res_file_row_height_bounded(self):
        """_shelx_res_file is truncated to MAX_DISPLAY_LINES; row should be medium height."""
        row = self.table.row_from_key('_shelx_res_file')
        h = self.table.rowHeight(row)
        self.assertGreaterEqual(h, 30)
        self.assertLessEqual(h, 500)

    def test_shelx_hkl_file_row_height_bounded(self):
        """_shelx_hkl_file is very large but truncated; row height must stay bounded."""
        row = self.table.row_from_key('_shelx_hkl_file')
        h = self.table.rowHeight(row)
        self.assertLessEqual(h, 600)

    def test_empty_value_row_stays_compact(self):
        """A row whose CIF value is '?' and has no other data should be compact."""
        row = self.table.row_from_key('_chemical_name_common')
        h = self.table.rowHeight(row)
        self.assertGreaterEqual(h, 10)
        self.assertLessEqual(h, 80)

    def test_row_height_after_equipment_load(self):
        """After loading equipment data the row should still have a reasonable height."""
        self.app.equipment.import_equipment_from_file(
            str(data.parent / 'test-data/Crystallographer_Details.cif'))
        self.app.ui.EquipmentTemplatesStackedWidget.setCurrentIndex(0)
        item = self.app.ui.EquipmentTemplatesListWidget.findItems(
            'Crystallographer Details', Qt.MatchStartsWith)[0]
        self.app.ui.EquipmentTemplatesListWidget.setCurrentItem(item)
        self.app.equipment.load_selected_equipment()
        self.table.resizeRowsToContents()
        row = self.table.row_from_key('_audit_contact_author_email')
        h = self.table.rowHeight(row)
        self.assertLessEqual(h, 100)


# ---------------------------------------------------------------------------
# Isolated tests – no AppWindow, just CifTableModel / CifTableView directly.
# ---------------------------------------------------------------------------

class TestCifTableModelIsolated(unittest.TestCase):
    """Tests for CifTableModel that cover previously untested branches."""

    def _make_model(self) -> CifTableModel:
        model = CifTableModel()
        model.bulk_load([
            CifRowData(key='_key_a', cif_value='val_a', data_value='d_a', edit_value='e_a'),
            CifRowData(key='_key_sep', is_separator=True),
            CifRowData(key='_key_c', cif_value='val_c',
                       cif_color=QColor(255, 0, 0),
                       data_color=QColor(0, 255, 0),
                       edit_color=QColor(0, 0, 255)),
        ])
        return model

    # --- data() ---

    def test_data_invalid_index_returns_none(self):
        model = self._make_model()
        self.assertIsNone(model.data(QModelIndex(), Qt.ItemDataRole.DisplayRole))

    def test_data_separator_background_role_returns_brush(self):
        model = self._make_model()
        result = model.data(model.index(1, 0), Qt.ItemDataRole.BackgroundRole)
        self.assertIsInstance(result, QBrush)

    def test_data_separator_display_role_returns_empty_string(self):
        model = self._make_model()
        self.assertEqual('', model.data(model.index(1, 0), Qt.ItemDataRole.DisplayRole))

    def test_data_separator_edit_role_returns_empty_string(self):
        model = self._make_model()
        self.assertEqual('', model.data(model.index(1, 0), Qt.ItemDataRole.EditRole))

    def test_data_separator_other_role_returns_none(self):
        model = self._make_model()
        self.assertIsNone(model.data(model.index(1, 0), Qt.ItemDataRole.DecorationRole))

    # --- setData() ---

    def test_setdata_invalid_index_returns_false(self):
        model = self._make_model()
        self.assertFalse(model.setData(QModelIndex(), 'v', Qt.ItemDataRole.EditRole))

    def test_setdata_background_role_col0(self):
        model = self._make_model()
        color = QColor(1, 2, 3)
        self.assertTrue(model.setData(model.index(0, 0), color, Qt.ItemDataRole.BackgroundRole))
        self.assertEqual(color, model._rows[0].cif_color)

    def test_setdata_background_role_col1(self):
        model = self._make_model()
        color = QColor(4, 5, 6)
        self.assertTrue(model.setData(model.index(0, 1), color, Qt.ItemDataRole.BackgroundRole))
        self.assertEqual(color, model._rows[0].data_color)

    def test_setdata_background_role_col2(self):
        model = self._make_model()
        color = QColor(7, 8, 9)
        self.assertTrue(model.setData(model.index(0, 2), color, Qt.ItemDataRole.BackgroundRole))
        self.assertEqual(color, model._rows[0].edit_color)

    def test_setdata_unknown_role_returns_false(self):
        model = self._make_model()
        self.assertFalse(model.setData(model.index(0, 0), 'x', Qt.ItemDataRole.DecorationRole))

    # --- flags() ---

    def test_flags_invalid_index_returns_no_flags(self):
        model = self._make_model()
        self.assertEqual(Qt.ItemFlag.NoItemFlags, model.flags(QModelIndex()))

    def test_flags_separator_row_is_only_enabled(self):
        model = self._make_model()
        result = model.flags(model.index(1, 0))
        self.assertEqual(Qt.ItemFlag.ItemIsEnabled, result)

    def test_flags_regular_col0_not_editable(self):
        model = self._make_model()
        result = model.flags(model.index(0, 0))
        self.assertTrue(result & Qt.ItemFlag.ItemIsEnabled)
        self.assertTrue(result & Qt.ItemFlag.ItemIsSelectable)
        self.assertFalse(result & Qt.ItemFlag.ItemIsEditable)

    def test_flags_regular_col2_is_editable(self):
        model = self._make_model()
        result = model.flags(model.index(0, 2))
        self.assertTrue(result & Qt.ItemFlag.ItemIsEditable)

    # --- headerData() ---

    def test_header_data_vertical_background_separator_returns_brush(self):
        model = self._make_model()
        result = model.headerData(1, Qt.Orientation.Vertical, Qt.ItemDataRole.BackgroundRole)
        self.assertIsInstance(result, QBrush)

    def test_header_data_vertical_background_non_separator_returns_none(self):
        model = self._make_model()
        result = model.headerData(0, Qt.Orientation.Vertical, Qt.ItemDataRole.BackgroundRole)
        self.assertIsNone(result)

    # --- row_from_key() ---

    def test_row_from_key_missing_raises_value_error(self):
        model = self._make_model()
        with self.assertRaises(ValueError):
            model.row_from_key('_nonexistent_key')

    # --- remove_row() ---

    def test_remove_row_out_of_range_returns_empty_string(self):
        model = self._make_model()
        self.assertEqual('', model.remove_row(999))

    def test_remove_row_negative_returns_empty_string(self):
        model = self._make_model()
        self.assertEqual('', model.remove_row(-1))

    # --- get_row_data() ---

    def test_get_row_data_out_of_bounds_returns_none(self):
        model = self._make_model()
        self.assertIsNone(model.get_row_data(999))

    def test_get_row_data_negative_returns_none(self):
        model = self._make_model()
        self.assertIsNone(model.get_row_data(-1))


class TestCifTableViewIsolated(unittest.TestCase):
    """Tests for CifTableView that cover previously untested branches."""

    def _make_view(self) -> CifTableView:
        view = CifTableView()
        view.bulk_load([
            CifRowData(key='_key_a', cif_value='val_a', data_value='d_a', edit_value='e_a'),
            CifRowData(key='_key_b', cif_value='val_b', data_value='d_b', edit_value='e_b'),
            CifRowData(key='_key_c', cif_value='val_c', data_value='d_c', edit_value='e_c'),
        ])
        return view

    def setUp(self):
        self.view = self._make_view()

    def tearDown(self):
        self.view.close()

    # --- counts ---

    def test_columns_count(self):
        self.assertEqual(3, self.view.columns_count)

    # --- delete_row variants ---

    def test_delete_row_without_argument_uses_current_index(self):
        self.view.setCurrentCell(1, 0)
        count_before = self.view.rowCount()
        self.view.delete_row()
        self.assertEqual(count_before - 1, self.view.rowCount())

    def test_delete_row_out_of_range_is_silent(self):
        count_before = self.view.rowCount()
        self.view.delete_row(999)
        self.assertEqual(count_before, self.view.rowCount())

    def test_delete_current_row_via_private_method(self):
        self.view.setCurrentCell(0, 0)
        count_before = self.view.rowCount()
        self.view._delete_current_row()
        self.assertEqual(count_before - 1, self.view.rowCount())

    # --- navigation ---

    def test_set_current_cell_and_current_row(self):
        self.view.setCurrentCell(2, 1)
        self.assertEqual(2, self.view.currentRow())

    def test_goto_template_page_emits_text_template_signal(self):
        received = []
        self.view.textTemplate.connect(lambda r: received.append(r))
        self.view.goto_template_page(1)
        self.assertEqual([1], received)

    # --- search ---

    def test_search_empty_string_unhides_all_rows(self):
        self.view.search('_key_a')  # hide non-matching rows first
        self.view.search('')
        for row in range(self.view.rowCount()):
            self.assertFalse(self.view.isRowHidden(row))

    def test_search_filters_non_matching_rows(self):
        self.view.search('_key_a')
        self.assertFalse(self.view.isRowHidden(0))  # _key_a – matches
        self.assertTrue(self.view.isRowHidden(1))   # _key_b – hidden
        self.assertTrue(self.view.isRowHidden(2))   # _key_c – hidden

    # --- clipboard ---

    def test_copy_vhead_item_puts_key_in_clipboard(self):
        self.view.setCurrentCell(0, 0)
        self.view.copy_vhead_item()
        self.assertEqual('_key_a', QApplication.clipboard().text())

    # --- _vheader_section_click ---

    def test_vheader_click_out_of_range_does_not_call_help(self):
        with patch('finalcif.gui.cif_table_view.show_keyword_help') as mock_help:
            self.view._vheader_section_click(999)
        mock_help.assert_not_called()

    def test_vheader_click_unknown_key_shows_no_help_message(self):
        with patch('finalcif.gui.cif_table_view.show_keyword_help') as mock_help:
            self.view._vheader_section_click(0)  # _key_a not in cif_all_dict
        mock_help.assert_called_once()
        self.assertIn('No help available', mock_help.call_args[0][1])

    def test_vheader_click_vrf_key_shows_vrf_help(self):
        self.view.bulk_load([CifRowData(key='_vrf_alert_A', cif_value='')])
        with patch('finalcif.gui.cif_table_view.show_keyword_help') as mock_help:
            self.view._vheader_section_click(0)
        mock_help.assert_called_once()
        self.assertIn('Validation Response Form', mock_help.call_args[0][1])

    def test_vheader_click_known_cif_key_shows_dict_help(self):
        from finalcif.cif.all_cif_dicts import cif_all_dict
        known_key = next(iter(cif_all_dict))
        self.view.bulk_load([CifRowData(key=known_key, cif_value='')])
        with patch('finalcif.gui.cif_table_view.show_keyword_help') as mock_help:
            self.view._vheader_section_click(0)
        mock_help.assert_called_once()

    # --- eventFilter ---

    def test_event_filter_tab_key_navigates_to_edit_column(self):
        self.view.setCurrentCell(0, 0)
        event = QKeyEvent(QEvent.Type.KeyRelease, Qt.Key.Key_Tab, Qt.KeyboardModifier.NoModifier)
        result = self.view.eventFilter(self.view, event)
        self.assertTrue(result)
        self.assertEqual(2, self.view.currentIndex().column())

    def test_event_filter_backtab_key_goes_to_previous_row_edit_column(self):
        self.view.setCurrentCell(2, 0)
        event = QKeyEvent(QEvent.Type.KeyRelease, Qt.Key.Key_Backtab, Qt.KeyboardModifier.NoModifier)
        result = self.view.eventFilter(self.view, event)
        self.assertTrue(result)
        self.assertEqual(1, self.view.currentIndex().row())
        self.assertEqual(2, self.view.currentIndex().column())

    def test_event_filter_non_key_event_passes_through(self):
        event = QEvent(QEvent.Type.MouseButtonPress)
        result = self.view.eventFilter(self.view, event)
        self.assertFalse(result)

    # --- add_property_combobox TypeError ---

    def test_add_property_combobox_bad_value_does_not_raise(self):
        """TypeError from retranslate_delimiter(None) is caught; bad item is skipped."""
        bad_data = [(0, None)]
        # Should not propagate the TypeError; the bad item is silently skipped.
        self.view.add_property_combobox(bad_data, 0, '_key_a')
        combobox = self.view.cellWidget(0, Column.EDIT)
        # The bad item was skipped, so the combobox has no items from bad_data.
        self.assertEqual(0, combobox.count())

    def test_add_property_combobox_good_items_added_despite_bad_ones(self):
        """Good items are still added when bad ones are interspersed."""
        mixed_data = [(0, 'good_value'), (1, None), (2, 'another_good')]
        self.view.add_property_combobox(mixed_data, 0, '_key_a')
        combobox = self.view.cellWidget(0, Column.EDIT)
        self.assertEqual(2, combobox.count())



from pathlib import Path
from unittest.mock import Mock

from finalcif.cif.cif_file_io import CifContainer
from finalcif.gui.dialogs import show_general_warning
from finalcif.tools.statusbar import StatusBar


class ShredCIF:
    """
    This class extracts the .res and .hkl file content from a cif file.
    """

    def __init__(self, cif: CifContainer, statusbar: StatusBar | None) -> None:
        self._cif = cif
        self._statusbar = statusbar or Mock()

    def shred_cif(self) -> None:
        """
        Saves res and hkl file from the cif.
        """
        self._statusbar.show_message('')
        if self._cif.is_multi_cif:
            resfile_path = self._cif.finalcif_file_prefixed(prefix='',
                                                            suffix='_' + self._cif.block.name + '-finalcif.res')
            hklfile_path = self._cif.finalcif_file_prefixed(prefix='',
                                                            suffix='_' + self._cif.block.name + '-finalcif.hkl')
            fabfile_path = self._cif.finalcif_file_prefixed(prefix='',
                                                            suffix='_' + self._cif.block.name + '-finalcif.fab')
        else:
            resfile_path = self._cif.finalcif_file.with_suffix('.res')
            hklfile_path = self._cif.finalcif_file.with_suffix('.hkl')
            fabfile_path = self._cif.finalcif_file.with_suffix('.fab')
        res_data = None
        hkl_data = None
        if not self._cif:
            return
        if not any([self._cif.res_file_data, self._cif.hkl_file]):
            self._statusbar.show_message('No .res and .hkl file data found!')
            return
        if self._cif.res_file_data:
            res_data = self._cif.res_file_data
        if self._cif.hkl_file:
            hkl_data = self._cif.hkl_file
        if not self._data_is_valid(res_data):
            self._statusbar.show_message('No .res file data found!')
        elif not self._write_res_file(resfile_path, res_data):
            return None
        if not self._data_is_valid(hkl_data):
            self._statusbar.show_message('No .hkl file data found!')
        else:
            hkl_data = self.format_hkl_data(hkl_data)
            if not self._write_hkl_file(hklfile_path, hkl_data):
                return None
        if fab_data := self._cif.fab_file:
            self._write_hkl_file(fabfile_path, fab_data)
        self._show_info(resfile_path, hklfile_path, res_data, hkl_data)

    @staticmethod
    def format_hkl_data(hkl_data: str) -> str:
        """
        Replaces ')' characters at the start of lines in the hkl file (after the 0 0 0 reflection)
        with ';' characters. The round brackets are placed there by SADABS, because a semicolon
        would interfere with the CIF syntax.
        """
        lines = []
        for line in hkl_data.splitlines(keepends=False):
            if line.startswith(')'):
                line = ';' + line[1:]
            lines.append(line)
        return '\n'.join(lines).lstrip('\n')

    def _show_info(self, resname: Path, hklname: Path, resdata: str | None, hkldata: str | None) -> None:
        if resdata and not hkldata:
            self._statusbar.show_message(f'{self._statusbar.current_message}\nFinished writing data to {resname.name}.')
        if hkldata and not resdata:
            self._statusbar.show_message(f'{self._statusbar.current_message}\nFinished writing data to {hklname.name}.')
        if hkldata and resdata:
            self._statusbar.show_message(
                f'{self._statusbar.current_message}\nFinished writing data to {resname.name} and {hklname.name}.')

    @staticmethod
    def _data_is_valid(data: str | None) -> bool:
        if not data or len(data.splitlines(keepends=False)) < 3:
            return False
        else:
            return True

    @staticmethod
    def _write_hkl_file(hklfile: Path, hkl: str | None) -> bool:
        try:
            hklfile.write_text(hkl, encoding='latin1', errors='ignore')
        except Exception as e:
            print(e)
            show_general_warning(parent=None, warn_text='Unable to write files: ' + str(e))
            return False
        return True

    def cif_has_hkl_or_res_file(self) -> bool:
        """
        Check whether hkl and/or res file content is included in the cif file.
        """
        if not self._cif.res_file_data:
            self._statusbar.show_message('No .res file data found!')
        if not self._cif.hkl_file and isinstance(self._statusbar.current_message, str):
            self._statusbar.show_message(self._statusbar.current_message + '\nNo .hkl file data found!')
        if not any([self._cif.res_file_data, self._cif.hkl_file]):
            self._statusbar.show_message('No .res and .hkl file data found!')
            return False
        else:
            return True

    @staticmethod
    def _write_res_file(resfile: Path, reslines: str | None) -> bool:
        """
        Writes a res file from the cif content.
        """
        try:
            resfile.write_text(reslines, encoding='latin1', errors='ignore', newline='\n')
        except Exception as e:
            print(e)
            show_general_warning(parent=None, warn_text='Unable to write files: ' + str(e))
            return False
        return True



"""Tests for JSON-based export/import of all templates."""

import json
import pickle
import shutil
import tempfile
from pathlib import Path
from unittest import TestCase

from finalcif.tools.settings import _custom_encoder, _custom_decoder, load_template_file


def _make_sample_templates(include_author: bool = False) -> dict:
    """Build sample template data.

    When *include_author* is True the authors list contains an Author-tagged
    dict that will trigger the full import chain of
    ``finalcif.equip_property.author_loop_templates``.  Set it to False for
    lightweight tests that should work without the full dependency tree.
    """
    authors: list = []
    if include_author:
        authors = [{
            '__type__': 'Author',
            'name': 'Jane Doe',
            'address': 'University',
            'email': 'jane@example.com',
            'phone': '',
            'orcid': '0000-0000-0000-0001',
            'footnote': '',
            'contact_author': True,
            'author_type': 'publ',
            'iucr_id': '',
        }]
    return {
        'text': [{'cif_key': '_refine_special_details', 'data': ['some text']}],
        'equipment': [{'name': 'Test Equip', 'data': [['_diffrn_source', 'Mo']], 'deleted': []}],
        'properties': [{'name': 'Test Prop', 'cif_key': '_exptl_absorpt_correction_type', 'data': ['multi-scan']}],
        'authors': authors,
        'cif_order': ['_cell_length_a', '_cell_length_b'],
        'cif_order_essential': ['_cell_length_a'],
    }


def _write_json(filepath: Path, templates: dict) -> None:
    with open(filepath, 'w', encoding='utf-8') as f:
        json.dump(templates, f, default=_custom_encoder, ensure_ascii=False, indent=2)


class TestLoadTemplateFile(TestCase):
    """Unit tests for load_template_file (no GUI required)."""

    def setUp(self) -> None:
        self.tmpdir = Path(tempfile.mkdtemp())

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    # -- JSON round-trip ---------------------------------------------------

    def test_export_json_is_valid(self):
        """Exported JSON file should be parseable and contain expected keys."""
        filepath = self.tmpdir / 'templates.json'
        _write_json(filepath, _make_sample_templates())

        loaded = load_template_file(filepath)
        self.assertIsInstance(loaded, dict)
        for key in ('text', 'equipment', 'properties', 'authors', 'cif_order', 'cif_order_essential'):
            self.assertIn(key, loaded)

    def test_json_text_templates_roundtrip(self):
        filepath = self.tmpdir / 'templates.json'
        _write_json(filepath, _make_sample_templates())

        loaded = load_template_file(filepath)
        self.assertEqual([{'cif_key': '_refine_special_details', 'data': ['some text']}], loaded['text'])

    def test_json_equipment_roundtrip(self):
        filepath = self.tmpdir / 'templates.json'
        _write_json(filepath, _make_sample_templates())

        loaded = load_template_file(filepath)
        self.assertEqual('Test Equip', loaded['equipment'][0]['name'])
        self.assertEqual([['_diffrn_source', 'Mo']], loaded['equipment'][0]['data'])

    def test_json_properties_roundtrip(self):
        filepath = self.tmpdir / 'templates.json'
        _write_json(filepath, _make_sample_templates())

        loaded = load_template_file(filepath)
        self.assertEqual('Test Prop', loaded['properties'][0]['name'])
        self.assertEqual('_exptl_absorpt_correction_type', loaded['properties'][0]['cif_key'])

    def test_json_cif_order_roundtrip(self):
        filepath = self.tmpdir / 'templates.json'
        _write_json(filepath, _make_sample_templates())

        loaded = load_template_file(filepath)
        self.assertEqual(['_cell_length_a', '_cell_length_b'], loaded['cif_order'])
        self.assertEqual(['_cell_length_a'], loaded['cif_order_essential'])

    def test_json_authors_roundtrip(self):
        """Author dataclass should survive JSON round-trip via object_hook.

        This test requires the full finalcif dependency tree (gemmi, numpy,
        shelxfile, etc.) because the custom decoder imports Author from
        author_loop_templates.
        """
        try:
            from finalcif.equip_property.author_loop_templates import Author
        except ImportError:
            self.skipTest("Full dependency tree not installed")

        filepath = self.tmpdir / 'templates.json'
        _write_json(filepath, _make_sample_templates(include_author=True))

        loaded = load_template_file(filepath)
        author = loaded['authors'][0]
        self.assertNotIsInstance(author, dict)
        self.assertEqual('Jane Doe', author.name)
        self.assertTrue(author.contact_author)

    # -- Legacy pickle fallback -------------------------------------------

    def test_legacy_pickle_dat_file(self):
        """A legacy .dat pickle file should still be importable."""
        filepath = self.tmpdir / 'legacy.dat'
        templates = _make_sample_templates()
        with open(filepath, 'wb') as f:
            pickle.dump(templates, f)

        loaded = load_template_file(filepath)
        self.assertIsInstance(loaded, dict)
        self.assertIn('text', loaded)
        self.assertIn('authors', loaded)

    # -- .dat file that contains JSON (edge case) -------------------------

    def test_dat_file_with_json_content(self):
        """A .dat file that actually contains JSON should be loaded as JSON."""
        filepath = self.tmpdir / 'new_format.dat'
        _write_json(filepath, _make_sample_templates())

        loaded = load_template_file(filepath)
        self.assertIsInstance(loaded, dict)
        self.assertEqual('Test Equip', loaded['equipment'][0]['name'])



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'text_templates_ui.ui'
##
##
################################################################################

from qtpy.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from qtpy.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from qtpy.QtWidgets import (QAbstractItemView, QApplication, QFrame, QGroupBox,
    QHBoxLayout, QLabel, QLineEdit, QListView,
    QListWidget, QListWidgetItem, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

from finalcif.gui.spell_check_edit import SpellTextEdit

class Ui_TextTemplatesWidget(object):
    def setupUi(self, TextTemplatesWidget):
        if not TextTemplatesWidget.objectName():
            TextTemplatesWidget.setObjectName(u"TextTemplatesWidget")
        TextTemplatesWidget.resize(883, 610)
        self.verticalLayout_3 = QVBoxLayout(TextTemplatesWidget)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(6, 6, 6, 6)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(6, -1, -1, 6)
        self.cifKeyLabel = QLabel(TextTemplatesWidget)
        self.cifKeyLabel.setObjectName(u"cifKeyLabel")

        self.horizontalLayout.addWidget(self.cifKeyLabel)

        self.cifKeyLineEdit = QLineEdit(TextTemplatesWidget)
        self.cifKeyLineEdit.setObjectName(u"cifKeyLineEdit")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(40)
        sizePolicy.setHeightForWidth(self.cifKeyLineEdit.sizePolicy().hasHeightForWidth())
        self.cifKeyLineEdit.setSizePolicy(sizePolicy)
        font = QFont()
        font.setBold(True)
        self.cifKeyLineEdit.setFont(font)
        self.cifKeyLineEdit.setReadOnly(True)

        self.horizontalLayout.addWidget(self.cifKeyLineEdit)

        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_3)


        self.verticalLayout_3.addLayout(self.horizontalLayout)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(-1, 0, -1, -1)
        self.TemplatesListGroupBox = QGroupBox(TextTemplatesWidget)
        self.TemplatesListGroupBox.setObjectName(u"TemplatesListGroupBox")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(50)
        sizePolicy1.setHeightForWidth(self.TemplatesListGroupBox.sizePolicy().hasHeightForWidth())
        self.TemplatesListGroupBox.setSizePolicy(sizePolicy1)
        self.verticalLayout = QVBoxLayout(self.TemplatesListGroupBox)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(-1, 6, -1, 6)
        self.templatesListWidget = QListWidget(self.TemplatesListGroupBox)
        self.templatesListWidget.setObjectName(u"templatesListWidget")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(60)
        sizePolicy2.setHeightForWidth(self.templatesListWidget.sizePolicy().hasHeightForWidth())
        self.templatesListWidget.setSizePolicy(sizePolicy2)
        self.templatesListWidget.setFrameShape(QFrame.NoFrame)
        self.templatesListWidget.setFrameShadow(QFrame.Plain)
        self.templatesListWidget.setLineWidth(0)
        self.templatesListWidget.setAutoScroll(False)
        self.templatesListWidget.setSelectionMode(QAbstractItemView.NoSelection)
        self.templatesListWidget.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.templatesListWidget.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.templatesListWidget.setResizeMode(QListView.Adjust)

        self.verticalLayout.addWidget(self.templatesListWidget)


        self.horizontalLayout_2.addWidget(self.TemplatesListGroupBox)

        self.combinedTestGroupBox = QGroupBox(TextTemplatesWidget)
        self.combinedTestGroupBox.setObjectName(u"combinedTestGroupBox")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(40)
        sizePolicy3.setHeightForWidth(self.combinedTestGroupBox.sizePolicy().hasHeightForWidth())
        self.combinedTestGroupBox.setSizePolicy(sizePolicy3)
        self.verticalLayout_4 = QVBoxLayout(self.combinedTestGroupBox)
        self.verticalLayout_4.setSpacing(6)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(12, 6, 12, 6)
        self.plainTextEdit = SpellTextEdit(self.combinedTestGroupBox)
        self.plainTextEdit.setObjectName(u"plainTextEdit")
        sizePolicy4 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.plainTextEdit.sizePolicy().hasHeightForWidth())
        self.plainTextEdit.setSizePolicy(sizePolicy4)
        self.plainTextEdit.setFrameShadow(QFrame.Plain)
        self.plainTextEdit.setLineWidth(0)
        self.plainTextEdit.setBackgroundVisible(False)

        self.verticalLayout_4.addWidget(self.plainTextEdit)


        self.horizontalLayout_2.addWidget(self.combinedTestGroupBox)


        self.verticalLayout_3.addLayout(self.horizontalLayout_2)

        self.applaTextHorizontalLayout = QHBoxLayout()
        self.applaTextHorizontalLayout.setObjectName(u"applaTextHorizontalLayout")
        self.applyTextPushButton = QPushButton(TextTemplatesWidget)
        self.applyTextPushButton.setObjectName(u"applyTextPushButton")

        self.applaTextHorizontalLayout.addWidget(self.applyTextPushButton)

        self.cancelTextPushButton = QPushButton(TextTemplatesWidget)
        self.cancelTextPushButton.setObjectName(u"cancelTextPushButton")

        self.applaTextHorizontalLayout.addWidget(self.cancelTextPushButton)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.applaTextHorizontalLayout.addItem(self.horizontalSpacer)

        self.savePushButton = QPushButton(TextTemplatesWidget)
        self.savePushButton.setObjectName(u"savePushButton")

        self.applaTextHorizontalLayout.addWidget(self.savePushButton)

        self.importPushButton = QPushButton(TextTemplatesWidget)
        self.importPushButton.setObjectName(u"importPushButton")

        self.applaTextHorizontalLayout.addWidget(self.importPushButton)

        self.exportTextPushButton = QPushButton(TextTemplatesWidget)
        self.exportTextPushButton.setObjectName(u"exportTextPushButton")

        self.applaTextHorizontalLayout.addWidget(self.exportTextPushButton)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.applaTextHorizontalLayout.addItem(self.horizontalSpacer_2)

        self.deletePushButton = QPushButton(TextTemplatesWidget)
        self.deletePushButton.setObjectName(u"deletePushButton")

        self.applaTextHorizontalLayout.addWidget(self.deletePushButton)


        self.verticalLayout_3.addLayout(self.applaTextHorizontalLayout)


        self.retranslateUi(TextTemplatesWidget)
    # setupUi

    def retranslateUi(self, TextTemplatesWidget):
        TextTemplatesWidget.setWindowTitle(QCoreApplication.translate("TextTemplatesWidget", u"TextTemplatesWidget", None))
        self.cifKeyLabel.setText(QCoreApplication.translate("TextTemplatesWidget", u"CIF key:", None))
        self.cifKeyLineEdit.setText("")
        self.TemplatesListGroupBox.setTitle(QCoreApplication.translate("TextTemplatesWidget", u"Select template(s)", None))
        self.combinedTestGroupBox.setTitle(QCoreApplication.translate("TextTemplatesWidget", u"Combined text", None))
        self.plainTextEdit.setPlaceholderText(QCoreApplication.translate("TextTemplatesWidget", u"The text of the selected templates is added here in the order of selection.", None))
        self.applyTextPushButton.setText(QCoreApplication.translate("TextTemplatesWidget", u"Apply Text", None))
        self.cancelTextPushButton.setText(QCoreApplication.translate("TextTemplatesWidget", u"Cancel", None))
        self.savePushButton.setText(QCoreApplication.translate("TextTemplatesWidget", u"Save as Template", None))
        self.importPushButton.setText(QCoreApplication.translate("TextTemplatesWidget", u"Import Template", None))
        self.exportTextPushButton.setText(QCoreApplication.translate("TextTemplatesWidget", u"Export to CIF", None))
        self.deletePushButton.setText(QCoreApplication.translate("TextTemplatesWidget", u"Delete Template", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'import_selector_ui.ui'
##
##
################################################################################

from qtpy.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from qtpy.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from qtpy.QtWidgets import (QApplication, QHBoxLayout, QHeaderView, QLabel,
    QMainWindow, QPushButton, QSizePolicy, QSpacerItem,
    QTableWidget, QTableWidgetItem, QVBoxLayout, QWidget)

class Ui_importSelectMainWindow(object):
    def setupUi(self, importSelectMainWindow):
        if not importSelectMainWindow.objectName():
            importSelectMainWindow.setObjectName(u"importSelectMainWindow")
        importSelectMainWindow.resize(800, 600)
        self.centralwidget = QWidget(importSelectMainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.importInfoLabel = QLabel(self.centralwidget)
        self.importInfoLabel.setObjectName(u"importInfoLabel")

        self.verticalLayout.addWidget(self.importInfoLabel)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(-1, 10, 10, -1)
        self.importTable_keys = QTableWidget(self.centralwidget)
        if (self.importTable_keys.columnCount() < 1):
            self.importTable_keys.setColumnCount(1)
        __qtablewidgetitem = QTableWidgetItem()
        self.importTable_keys.setHorizontalHeaderItem(0, __qtablewidgetitem)
        self.importTable_keys.setObjectName(u"importTable_keys")
        self.importTable_keys.verticalHeader().setVisible(False)

        self.horizontalLayout_2.addWidget(self.importTable_keys)

        self.importTable_loops = QTableWidget(self.centralwidget)
        if (self.importTable_loops.columnCount() < 1):
            self.importTable_loops.setColumnCount(1)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.importTable_loops.setHorizontalHeaderItem(0, __qtablewidgetitem1)
        self.importTable_loops.setObjectName(u"importTable_loops")
        self.importTable_loops.verticalHeader().setVisible(False)

        self.horizontalLayout_2.addWidget(self.importTable_loops)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.importSelectedPushbutton = QPushButton(self.centralwidget)
        self.importSelectedPushbutton.setObjectName(u"importSelectedPushbutton")

        self.horizontalLayout.addWidget(self.importSelectedPushbutton)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.selectOnlyNewPB = QPushButton(self.centralwidget)
        self.selectOnlyNewPB.setObjectName(u"selectOnlyNewPB")

        self.horizontalLayout.addWidget(self.selectOnlyNewPB)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self.saveSelectionPushbutton = QPushButton(self.centralwidget)
        self.saveSelectionPushbutton.setObjectName(u"saveSelectionPushbutton")

        self.horizontalLayout.addWidget(self.saveSelectionPushbutton)


        self.verticalLayout.addLayout(self.horizontalLayout)

        importSelectMainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(importSelectMainWindow)
    # setupUi

    def retranslateUi(self, importSelectMainWindow):
        importSelectMainWindow.setWindowTitle(QCoreApplication.translate("importSelectMainWindow", u"Select import items", None))
        self.importInfoLabel.setText("")
        ___qtablewidgetitem = self.importTable_keys.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("importSelectMainWindow", u"keys", None))
        ___qtablewidgetitem1 = self.importTable_loops.horizontalHeaderItem(0)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("importSelectMainWindow", u"loops", None))
        self.importSelectedPushbutton.setText(QCoreApplication.translate("importSelectMainWindow", u"Import Selected", None))
        self.selectOnlyNewPB.setText(QCoreApplication.translate("importSelectMainWindow", u"Select Only New Data", None))
        self.saveSelectionPushbutton.setText(QCoreApplication.translate("importSelectMainWindow", u"Save current Selection", None))
    # retranslateUi




from finalcif.tools import chemparse


def formula_str_to_dict(sumform: str | bytes) -> dict[str, float]:
    chemical_formula = chemparse.parse_formula(sumform.replace(" ", ""))
    return chemical_formula


def sum_formula_to_html(sumform: dict[str, float | int], break_after: int = 99) -> str:
    """
    Makes html formatted sum formula from a dictionary.
    """
    if not sumform:
        return ''
    formula_list = ['<html><body>']
    num = 0
    for el, count in sumform.items():
        if count == 0 or count is None:
            continue
        try:
            times = round(float(count), 1)
        except (TypeError, ValueError):
            times = 1
        if num > 3 and num % break_after == 0:
            formula_list.append("<br>")
        if times == 1:
            formula_list.append(f'{el}')
        else:
            formula_list.append(f"{el}<sub>{times:g}</sub>")
        num += 1
    formula_list.append('</body></html>')
    formula = "".join(formula_list)
    return formula



#!/usr/bin/env python3
"""QPlainTextEdit With Inline Spell Check
Original PyQt4 Version:
    https://nachtimwald.com/2009/08/22/qplaintextedit-with-in-line-spell-check/
Copyright 2009 John Schember
Copyright 2018 Stephan Sokolow
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__license__ = 'MIT'
__author__ = 'John Schember; Stephan Sokolow'
__docformat__ = 'restructuredtext en'

import os
import sys
from typing import Union

from qtpy import QtGui, QtCore
from qtpy.QtCore import QEvent, QPoint
from qtpy.QtGui import (QFocusEvent, QSyntaxHighlighter, QTextBlockUserData, QTextCharFormat, QTextCursor,
                           QContextMenuEvent)
from qtpy.QtWidgets import (QApplication, QMenu, QPlainTextEdit)

try:
    if "RUNNING_TEST" in os.environ:
        raise ImportError
    import enchant
    from enchant import tokenize
    from enchant.errors import TokenizerNotFoundError
    from enchant.utils import trim_suggestions


    class SpellTextEdit(QPlainTextEdit):
        """QPlainTextEdit subclass which does spell-checking using PyEnchant"""

        # Clamping value for words like "regex" which suggest so many things that
        # the menu runs from the top to the bottom of the screen and spills over
        # into a second column.
        max_suggestions = 10

        def __init__(self, *args, **kwargs) -> None:
            QPlainTextEdit.__init__(self, *args, **kwargs)
            start_language = 'en_US'
            # Start with a default dictionary based on US english.
            self.highlighter = EnchantHighlighter(self.document())
            self.highlighter.setDict(enchant.Dict(start_language))

        def contextMenuEvent(self, event: QContextMenuEvent) -> None:
            """Custom context menu handler to add a spelling suggestions submenu"""
            popup_menu = self.create_spellcheck_context_menu(event.pos())
            popup_menu.exec(event.globalPos())

            # Fix bug observed in Qt 5.2.1 on *buntu 14.04 LTS where:
            # 1. The cursor remains invisible after closing the context menu
            # 2. Keyboard input causes it to appear, but it doesn't blink
            # 3. Switching focus away from and back to the window fixes it
            self.focusInEvent(QFocusEvent(QEvent.Type.FocusIn))

        def create_spellcheck_context_menu(self, pos: QPoint) -> QMenu:
            """Create and return an augmented default context menu.
            This may be used as an alternative to the QPoint-taking form of
            ``createStandardContextMenu`` and will work on pre-5.5 Qt.
            """
            try:  # Recommended for Qt 5.5+ (Allows contextual Qt-provided entries)
                menu = self.createStandardContextMenu(pos)
            except TypeError:  # Before Qt 5.5
                menu = self.createStandardContextMenu()
            menu.setTitle('Spell Checking')
            menu.insertMenu(menu.actions()[0], self.create_languages_menu(menu))
            # menu.insertSeparator(menu.actions()[1])
            # Add a submenu for setting the spell-check language
            cursor = self.cursor_for_misspelling(pos)
            self.create_learn_word_menu(cursor, menu)
            # menu.addMenu(self.createFormatsMenu(menu))
            # Try to retrieve a menu of corrections for the right-clicked word
            spell_menu = self.create_corrections_menu(cursor, menu)
            if spell_menu:
                menu.insertSeparator(menu.actions()[2])
                menu.insertMenu(menu.actions()[0], spell_menu)
            return menu

        def create_learn_word_menu(self, cursor, menu):
            if not cursor:
                return None
            text = cursor.selectedText()
            learn_action = QtGui.QAction(f"Learn '{text}'", parent=menu)
            learn_action.setData(text)
            menu.triggered.connect(self.learn_word)
            menu.insertAction(menu.actions()[0], learn_action)

        def learn_word(self, action: QtGui.QAction) -> None:
            misspelled_word = action.data()
            if 'Learn' not in action.text() or isinstance(misspelled_word, tuple):
                return None
            # Add the misspelled word to the personal dictionary
            personal_dictionary = self.highlighter.dict()
            personal_dictionary.add(misspelled_word)
            self.highlighter.rehighlight()

        def create_corrections_menu(self, cursor, parent=None) -> QMenu | None:
            """Create and return a menu for correcting the selected word."""
            if not cursor:
                return None

            text = cursor.selectedText()
            suggests = trim_suggestions(text, self.highlighter.dict().suggest(text), self.max_suggestions)

            spell_menu = QMenu('Spelling Suggestions', parent)
            for word in suggests:
                action = QtGui.QAction(word, spell_menu)
                action.setData((cursor, word))
                spell_menu.addAction(action)

            # Only return the menu if it's non-empty
            if spell_menu.actions():
                spell_menu.triggered.connect(self.cb_correct_word)
                return spell_menu
            return None

        def create_languages_menu(self, parent=None):
            """Create and return a menu for selecting the spell-check language."""
            curr_lang = self.highlighter.dict().tag
            lang_menu = QMenu("Language", parent)
            lang_actions = QtGui.QActionGroup(lang_menu)

            for lang in enchant.list_languages():
                action = lang_actions.addAction(lang)
                action.setCheckable(True)
                action.setChecked(lang == curr_lang)
                action.setData(lang)
                lang_menu.addAction(action)

            lang_menu.triggered.connect(self.cb_set_language)
            return lang_menu

        def create_formats_menu(self, parent=None):
            """Create and return a menu for selecting the spell-check language."""
            fmt_menu = QMenu("Format", parent)
            fmt_actions = QtGui.QActionGroup(fmt_menu)

            curr_format = self.highlighter.chunkers()
            for name, chunkers in (('Text', []), ('HTML', [tokenize.HTMLChunker])):
                action = fmt_actions.addAction(name)
                action.setCheckable(True)
                action.setChecked(chunkers == curr_format)
                action.setData(chunkers)
                fmt_menu.addAction(action)

            fmt_menu.triggered.connect(self.cb_set_format)
            return fmt_menu

        def cursor_for_misspelling(self, pos):
            """Return a cursor selecting the misspelled word at ``pos`` or ``None``
            This leverages the fact that QPlainTextEdit already has a system for
            processing its contents in limited-size blocks to keep things fast.
            """
            cursor = self.cursorForPosition(pos)
            misspelled_words = getattr(cursor.block().userData(), 'misspelled', [])

            # If the cursor is within a misspelling, select the word
            for (start, end) in misspelled_words:
                if start <= cursor.positionInBlock() <= end:
                    block_pos = cursor.block().position()

                    cursor.setPosition(block_pos + start, QTextCursor.MoveMode.MoveAnchor)
                    cursor.setPosition(block_pos + end, QTextCursor.MoveMode.KeepAnchor)
                    break

            if cursor.hasSelection():
                return cursor
            else:
                return None

        def cb_correct_word(self, action):
            """Event handler for 'Spelling Suggestions' entries."""
            cursor, word = action.data()
            cursor.beginEditBlock()
            cursor.removeSelectedText()
            cursor.insertText(word)
            cursor.endEditBlock()

        def cb_set_language(self, action):
            """Event handler for 'Language' menu entries."""
            lang = action.data()
            self.highlighter.setDict(enchant.Dict(lang))

        def cb_set_format(self, action):
            """Event handler for 'Language' menu entries."""
            chunkers = action.data()
            self.highlighter.setChunkers(chunkers)
            # TODO: Emit an event so this menu can trigger other things


    class CustomFilter(tokenize.Filter):
        r"""Filter skipping over listed words.
        """
        # _pattern = re.compile(r"^([A-Z]\w+[A-Z]+\w+)")
        words_to_skip = (
            'finalcif',
            'structurefinder',
            'ccdc',
            'cod',
            'shelxle',
            'shelx',
            'shelxl',
            'shelxd',
            'shelxt',
            'shelxs',
            'wingx',
        )

        def _skip(self, word) -> bool:
            if word.lower() in self.words_to_skip:
                return True
            return False


    class EnchantHighlighter(QSyntaxHighlighter):
        """QSyntaxHighlighter subclass which consults a PyEnchant dictionary"""
        tokenizer = None
        token_filters = (tokenize.EmailFilter, tokenize.URLFilter, CustomFilter)

        # Define the spellcheck style once and just assign it as necessary
        # XXX: Does QSyntaxHighlighter.setFormat handle keeping this from
        #      clobbering styles set in the data itself?
        err_format = QTextCharFormat()
        err_format.setUnderlineColor(QtCore.Qt.GlobalColor.red)
        # err_format.setUnderlineStyle(QTextCharFormat.SpellCheckUnderline)
        err_format.setUnderlineStyle(QTextCharFormat.UnderlineStyle.WaveUnderline)

        def __init__(self, *args) -> None:
            QSyntaxHighlighter.__init__(self, *args)

            # Initialize private members
            self._sp_dict = None
            self._chunkers = []

        def chunkers(self):
            """Gets the chunkers in use"""
            return self._chunkers

        def dict(self) -> enchant.Dict:
            """Gets the spelling dictionary in use"""
            return self._sp_dict

        def setChunkers(self, chunkers):
            """Sets the list of chunkers to be used"""
            self._chunkers = chunkers
            self.setDict(self.dict())
            # FIXME: Revert self._chunkers on failure to ensure consistent state

        def setDict(self, sp_dict):
            """Sets the spelling dictionary to be used"""
            try:
                self.tokenizer = tokenize.get_tokenizer(sp_dict.tag,
                                                        chunkers=self._chunkers, filters=self.token_filters)
            except TokenizerNotFoundError:
                # Fall back to the "good for most euro languages" English tokenizer
                self.tokenizer = tokenize.get_tokenizer(
                    chunkers=self._chunkers, filters=self.token_filters)
            self._sp_dict = sp_dict

            self.rehighlight()

        def highlightBlock(self, text: str) -> None:
            """Overridden QSyntaxHighlighter method to apply the highlight"""
            if not self._sp_dict:
                return

            # Build a list of all misspelled words and highlight them
            misspellings = []
            for (word, pos) in self.tokenizer(text):
                if not self._sp_dict.check(word):
                    self.setFormat(pos, len(word), self.err_format)
                    misspellings.append((pos, pos + len(word)))

            # Store the list so the context menu can reuse this tokenization pass
            # (Block-relative values so editing other blocks won't invalidate them)
            data = QTextBlockUserData()
            data.misspelled = misspellings
            self.setCurrentBlockUserData(data)
except ImportError:
    class SpellTextEdit(QPlainTextEdit):
        pass

if __name__ == '__main__':
    app = QApplication(sys.argv)

    spellEdit = SpellTextEdit()
    spellEdit.show()
    spellEdit.setPlainText('This is somee missspelled texawt with errorsd.')

    sys.exit(app.exec())



import sys
import zipfile
from pathlib import Path

from finalcif.gui.custom_classes import light_red
from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import Qt


class VZSImageViewer(QtWidgets.QWidget):
    zip_path: None

    def __init__(self, parent=None):
        super().__init__(parent)
        self.zipfile: zipfile.ZipFile
        self.image_names: list[Path] | None = None
        self.index = 0
        self.pixmap = None
        # self.setMinimumSize(300, 300)
        self.zoom_rect = None
        self.zoomed = False
        self.zoom_start = None
        self.zoom_end = None
        self.rotation: float = 0.0

    def load_file(self, file_path: Path) -> None:
        if not file_path or not file_path.is_file():
            return
        if file_path.suffix.lower() == '.vzs':
            self._load_zip(file_path)
        elif file_path.suffix.lower() == '.jpg':
            self.load_jpegs(file_path)

    def load_jpegs(self, file_path: Path):
        self.zipfile = None
        globber = f'{file_path.stem.rstrip("1234567890")}*.jpg'
        self.image_names = list(file_path.parent.glob(globber))
        self.image_names.sort()
        self._load_current_image()

    def _load_zip(self, zip_path: Path) -> None:
        self.zipfile = zipfile.ZipFile(zip_path, mode="r")
        if self.zipfile.namelist():
            self.image_names = [name for name in self.zipfile.namelist()
                                if name.lower().endswith(".jpg")]
            self.image_names.sort()
            self._load_orientation()
            self._load_current_image()

    def _load_orientation(self) -> None:
        parameters = 'params.txt'
        if parameters in self.zipfile.namelist():
            for line in self.zipfile.open(parameters):
                if line.startswith(b'Microscope Rotation'):
                    splitline = line.decode('latin1').rstrip().split('=')
                    if len(splitline) > 1:
                        _, rotation = splitline
                        try:
                            self.rotation = float(rotation)
                        except ValueError:
                            print(f'Rotation value invalid: {rotation}')
                            self.rotation = 0.0
                        return None

    def reset(self) -> None:
        self.pixmap = None
        self.update()

    def _load_current_image(self) -> None:
        if not self.image_names:
            self.pixmap = None
            return
        data = self._read_data()
        image = QtGui.QImage()
        image.loadFromData(data)
        t = QtGui.QTransform().rotate(self.rotation)
        rotated = image.transformed(t, QtCore.Qt.TransformationMode.SmoothTransformation)
        self.pixmap = QtGui.QPixmap().fromImage(rotated)
        self.update()

    def save_image(self, file_path: Path) -> None:
        size = self.pixmap.size()
        cropped = self.pixmap.copy(self.zoom_rect)
        scaled = cropped.scaled(size, Qt.AspectRatioMode.KeepAspectRatio,
                                Qt.TransformationMode.SmoothTransformation)
        self.pixmap = scaled
        self.pixmap.save(str(file_path))

    def _read_data(self) -> bytes:
        if self.zipfile:
            with self.zipfile.open(self.image_names[self.index]) as f:
                data = f.read()
        else:
            with open(self.image_names[self.index], 'rb') as f:
                data = f.read()
        return data

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        if not self.pixmap:
            painter.drawText(self.rect(), Qt.AlignmentFlag.AlignCenter,
                             "No images found.\nSelect a crystal video file.")
            return
        if self.zoomed and self.zoom_rect:
            cropped = self.pixmap.copy(self.zoom_rect)
            scaled = cropped.scaled(self.size(), Qt.AspectRatioMode.KeepAspectRatio,
                                    Qt.TransformationMode.SmoothTransformation)
            painter.drawPixmap(self.rect().center() - scaled.rect().center(), scaled)
        else:
            scaled = self.pixmap.scaled(self.size(), Qt.AspectRatioMode.KeepAspectRatio,
                                        Qt.TransformationMode.SmoothTransformation)
            painter.drawPixmap(self.rect().center() - scaled.rect().center(), scaled)
        if self.zoom_start and self.zoom_end:
            pen = painter.pen()
            pen.setColor(light_red)
            pen.setWidth(1)
            painter.setPen(pen)
            rect = QtCore.QRectF(self.zoom_start, self.zoom_end)
            painter.drawRect(rect)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if not self.image_names:
            return
        if event.angleDelta().y() > 0:
            self.next()
        else:
            self.previous()

    def previous(self) -> None:
        self.index = (self.index + 1) % len(self.image_names)
        self._load_current_image()

    def next(self) -> None:
        self.index = (self.index - 1) % len(self.image_names)
        self._load_current_image()

    def mousePressEvent(self, event: QtGui.QMouseEvent):
        if event.button() == Qt.MouseButton.LeftButton and self.pixmap:
            self.zoom_start = event.position().toPoint()
            self.zoom_end = None

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        if self.zoom_start:
            self.zoom_end = event.position().toPoint()
            self.update()

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        if self.zoom_start and self.pixmap:
            self.zoom_end = event.position().toPoint()
            rect = QtCore.QRectF(self.zoom_start, self.zoom_end).normalized()
            scaled = self.pixmap.scaled(self.size(),
                                        Qt.AspectRatioMode.KeepAspectRatio,
                                        Qt.TransformationMode.SmoothTransformation)
            offset = self.rect().center() - scaled.rect().center()
            x_ratio = self.pixmap.width() / scaled.width()
            y_ratio = self.pixmap.height() / scaled.height()
            img_rect = scaled.rect().translated(offset)
            intersected = rect.intersected(img_rect)
            if intersected.isValid():
                x1 = (intersected.left() - offset.x()) * x_ratio
                y1 = (intersected.top() - offset.y()) * y_ratio
                x2 = (intersected.right() - offset.x()) * x_ratio
                y2 = (intersected.bottom() - offset.y()) * y_ratio
                self.zoom_rect = QtCore.QRectF(x1, y1, x2 - x1, y2 - y1).toRect().intersected(self.pixmap.rect())
                if self.zoom_rect.width() > 10 and self.zoom_rect.height() > 10:
                    self.zoomed = True
            self.zoom_start = None
            self.zoom_end = None
            self.update()
        elif event.button() == Qt.MouseButton.RightButton:
            self.zoomed = False
            self.zoom_rect = None
            self.update()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    widget = VZSImageViewer()
    widget.load_file(Path("test-data/BB_GS3.vzs"))
    widget.setWindowTitle("VZS Viewer")
    widget.show()
    sys.exit(app.exec())



import gemmi.cif
from gemmi import cif

from finalcif.gui.dialogs import show_general_warning


def read_document_from_cif_file(filename: str) -> gemmi.cif.Document | None:
    doc: gemmi.cif.Document | None = None
    try:
        doc = cif.read_file(filename)
    except (RuntimeError, ValueError) as e:
        warning = "{}\n{}".format('This CIF is invalid:', str(e))
        if 'data_' in str(e):
            warning = f"{warning}\n\nA CIF needs to start with 'data_[some_name]'."
        show_general_warning(parent=None, warn_text=warning)
    except OSError as e:
        show_general_warning(parent=None, warn_text=f'Unable to open file {filename}:\n{e!s}')
    return doc



import os
import subprocess
import sys
import threading
import time
from contextlib import suppress
from pathlib import Path
from shutil import which

from qtpy import QtCore
from qtpy.QtWidgets import QApplication, QMainWindow, QPushButton, QVBoxLayout, QWidget, QLabel, \
    QPlainTextEdit


class PlatonRunner(QtCore.QObject):
    finished = QtCore.Signal(bool)
    formula = QtCore.Signal(str)
    tick = QtCore.Signal(str)

    def __init__(self, parent, output_widget: QPlainTextEdit, log_widget: QPlainTextEdit, cif_file: Path):
        super().__init__(parent)
        self.cif_file = cif_file.resolve().absolute()
        self.process = None
        self.is_stopped = False
        self._origdir = None
        self.output_widget = output_widget
        self.log_widget = log_widget
        self.formula_moiety = ''
        self.Z = ''
        self.chk_file_text = ''

    def run_process(self):
        self._origdir = os.curdir
        # os.chdir(self.cif_file.parent)
        self.formula_moiety = ''
        self.Z = ''
        self.process = QtCore.QProcess()
        self.output_widget.clear()
        threading.Thread(target=self._monitor_output_log, daemon=True).start()
        # self.process.readyReadStandardOutput.connect(self.on_ready_read)
        self.process.finished.connect(self._onfinished)
        self.process.setWorkingDirectory(str(self.cif_file.parent))
        self.cif_file.with_suffix('.chk').unlink(missing_ok=True)
        self.process.start(self.platon_exe, ["-U", str(self.cif_file.name)])

    def _onfinished(self) -> None:
        self._on_ready_read()
        # os.chdir(self._origdir)
        self._parse_chk_file()
        self.output_widget.setPlainText(self.chk_file_text)
        self.finished.emit(True)
        self.delete_orphaned_files()

    def _on_ready_read(self) -> None:
        output = self.process.readAllStandardOutput().data().decode()
        self.log_widget.appendPlainText(output)

    def _monitor_output_log(self) -> None:
        while not self.is_stopped:
            self.tick.emit('#')
            time.sleep(0.3)
            try:
                log_file = self.cif_file.with_suffix('.chk').read_text('latin1', errors='ignore')
                if 'Unresolved or to be Checked Issue' in log_file:
                    self._stop_program()
                if '! Congratulations !' in log_file:
                    self._stop_program()
            except FileNotFoundError:
                break

    def _stop_program(self) -> None:
        self.is_stopped = True
        # if self.process and self.process.state() == QProcess.Running:
        self.process.terminate()
        self.finished.emit(True)

    def _parse_chk_file(self) -> None:
        try:
            self.chk_file_text = self.cif_file.with_suffix('.chk').read_text(encoding='latin1', errors='ignore')
        except FileNotFoundError as e:
            print('CHK file not found:', e)
            self.chk_file_text = ''
        for num, line in enumerate(self.chk_file_text.splitlines(keepends=False)):
            if line.startswith('# MoietyFormula'):
                self.formula_moiety = ' '.join(line.split(' ')[2:])
                self.formula.emit(self.formula_moiety)
            if line.startswith('# Z'):
                self.Z = line[19:24].strip(' ')

    @property
    def platon_exe(self) -> str:
        special_platon = Path(__file__).resolve().parent.parent.parent.joinpath('platon/platon_special.exe')
        if special_platon.exists():
            return str(special_platon)
        if sys.platform.startswith('win'):
            in_pwt = r'C:\pwt\platon.exe'
        else:
            in_pwt = 'platon'
        if Path(in_pwt).exists():
            return in_pwt
        elif which('platon'):
            return which('platon')
        else:
            return 'platon'

    def kill(self):
        if sys.platform.startswith('win'):
            with suppress(FileNotFoundError):
                subprocess.run(["taskkill", "/f", "/im", "platon.exe"], check=False, shell=False)
        if sys.platform[:5] in ('linux', 'darwi'):
            with suppress(FileNotFoundError):
                subprocess.run(["killall", "platon"], check=False, shell=False)

    def delete_orphaned_files(self):
        # delete orphaned files:
        for ext in ['.ckf', '.fcf', '.def', '.lis', '.sar', '.ckf',
                    '.sum', '.hkp', '.pjn', '.bin', '.spf']:
            try:
                file = self.cif_file.resolve().with_suffix(ext)
                if file.stat().st_size < 100:
                    file.unlink(missing_ok=True)
                if file.suffix in ['.sar', '.spf', '.ckf']:
                    file.unlink(missing_ok=True)
            except FileNotFoundError:
                pass


class ProcessWidget(QWidget):
    def __init__(self):
        super().__init__()
        layout = QVBoxLayout()
        self.text_widget = QPlainTextEdit()
        self.log_widget = QPlainTextEdit()
        layout.addWidget(self.log_widget)
        layout.addWidget(self.text_widget)
        self.button = QPushButton("Run QProcess")
        layout.addWidget(self.button)
        self.time_label = QLabel()
        layout.addWidget(self.time_label)
        self.setLayout(layout)
        self.runner = PlatonRunner(parent=self, output_widget=self.text_widget, log_widget=self.log_widget,
                                   cif_file=Path("tests/examples/work/cu_BruecknerJK_153F40_0m.cif"))
        self.button.clicked.connect(lambda x: self.button.setDisabled(True))
        self.button.clicked.connect(lambda x: self.runner.run_process())
        self.button.clicked.connect(lambda x: self.log_widget.setPlainText('Running Platon'))
        self.runner.finished.connect(lambda x: self.button.setEnabled(True))

        # Only to show that the main thread works continuously:
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.update_time)
        self.timer.start(1000)

    def update_time(self):
        current_time = QtCore.QTime.currentTime()
        time_text = current_time.toString("hh:mm:ss")
        self.time_label.setText(f"Current Time: {time_text}")


if __name__ == '__main__':
    app = QApplication(sys.argv)
    window = QMainWindow()
    window.setWindowTitle("QProcess Example")
    window.setMinimumWidth(800)
    window.setMinimumHeight(600)

    process_widget = ProcessWidget()
    window.setCentralWidget(process_widget)

    window.show()

    sys.exit(app.exec())



import unittest
from pathlib import Path

from finalcif.cif.cif_file_io import CifContainer
from finalcif.tools.shred import ShredCIF
from finalcif.tools.statusbar import StatusBar

data = Path('.')

class TestShedCifWithData(unittest.TestCase):

    def setUp(self) -> None:
        self.cif = CifContainer(data / 'test-data/p21c.cif')
        self.shred = ShredCIF(self.cif, statusbar=StatusBar())
        self.outfile_hkl = self.cif.finalcif_file.with_suffix('.hkl')
        self.outfile_res = self.cif.finalcif_file.with_suffix('.res')

    def tearDown(self) -> None:
        self.outfile_hkl.unlink(missing_ok=True)
        self.outfile_res.unlink(missing_ok=True)

    def test_no_shred(self):
        self.assertEqual(self.outfile_hkl.exists(), False)
        self.assertEqual(self.outfile_res.exists(), False)
        self.assertEqual(self.shred._statusbar.current_message, '')

    def test_shred(self):
        self.shred.shred_cif()
        self.assertEqual(self.outfile_res.exists(), True)
        self.assertEqual(self.outfile_hkl.exists(), True)
        self.assertEqual('\nFinished writing data to p21c-finalcif.res and p21c-finalcif.hkl.',
                         self.shred._statusbar.current_message)
        # Make sure the extracted hkl does start with a reflection and not a newline:
        self.assertEqual('   1   0   0  323.11   10.61\n',
                         self.outfile_hkl.read_text(encoding='latin1')[:29])


class TestShedCifNoData(unittest.TestCase):

    def setUp(self) -> None:
        self.cif = CifContainer(data / 'test-data/1000007.cif')
        self.shred = ShredCIF(self.cif, statusbar=StatusBar())
        self.outfile_hkl = self.cif.finalcif_file.with_suffix('.hkl')
        self.outfile_res = self.cif.finalcif_file.with_suffix('.res')

    def tearDown(self) -> None:
        self.outfile_hkl.unlink(missing_ok=True)
        self.outfile_res.unlink(missing_ok=True)

    def test_no_shred(self):
        self.assertEqual(self.outfile_hkl.exists(), False)
        self.assertEqual(self.outfile_res.exists(), False)
        self.assertEqual('', self.shred._statusbar.current_message)

    def test_shred(self):
        self.shred.shred_cif()
        self.assertEqual(self.outfile_res.exists(), False)
        self.assertEqual(self.outfile_hkl.exists(), False)
        self.assertEqual('No .res and .hkl file data found!', self.shred._statusbar.current_message)


class TestExport(unittest.TestCase):

    def setUp(self) -> None:
        self.cif = CifContainer(data / 'tests/examples/work/cu_BruecknerJK_153F40_0m.cif')
        self.outfile_hkl = self.cif.finalcif_file.with_suffix('.hkl')
        self.outfile_res = self.cif.finalcif_file.with_suffix('.res')
        self.shred = ShredCIF(self.cif, statusbar=StatusBar())
        self.shred.shred_cif()

    def tearDown(self):
        self.outfile_hkl.unlink(missing_ok=True)
        self.outfile_res.unlink(missing_ok=True)

    def test_export_hkl(self):
        """
        Shredcif test
        """
        test_hkl_file = data / 'tests/examples/work/test_hkl_file.txt'
        self.assertEqual(test_hkl_file.read_text().splitlines(keepends=True),
                         self.outfile_hkl.read_text().splitlines(keepends=True))

    def test_export_res(self):
        test_res_file = data / 'tests/examples/work/test_res_file.txt'
        self.assertEqual(test_res_file.read_text(), self.outfile_res.read_text())
